@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-
 # Copyright 2018 The Kubernetes Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -16,11 +14,13 @@
 
 import base64
 import datetime
+import io
 import json
 import os
 import shutil
 import tempfile
 import unittest
+from collections import namedtuple
 
 import mock
 import yaml
@@ -29,11 +29,15 @@ from six import PY3, next
 from kubernetes.client import Configuration
 
 from .config_exception import ConfigException
-from .kube_config import (ENV_KUBECONFIG_PATH_SEPARATOR, ConfigNode,
-                          FileOrData, KubeConfigLoader, KubeConfigMerger,
-                          _cleanup_temp_files, _create_temp_file_with_content,
+from .dateutil import parse_rfc3339
+from .kube_config import (ENV_KUBECONFIG_PATH_SEPARATOR, CommandTokenSource,
+                          ConfigNode, FileOrData, KubeConfigLoader,
+                          KubeConfigMerger, _cleanup_temp_files,
+                          _create_temp_file_with_content,
+                          _get_kube_config_loader,
+                          _get_kube_config_loader_for_yaml_file,
                           list_kube_config_contexts, load_kube_config,
-                          new_client_from_config)
+                          load_kube_config_from_dict, new_client_from_config)
 
 BEARER_TOKEN_FORMAT = "Bearer %s"
 
@@ -130,6 +134,10 @@ TEST_OIDC_INVALID_PADDING_LENGTH = ".".join([
 
 TEST_OIDC_CA = _base64(TEST_CERTIFICATE_AUTH)
 
+TEST_AZURE_LOGIN = TEST_OIDC_LOGIN
+TEST_AZURE_TOKEN = "test-azure-token"
+TEST_AZURE_TOKEN_FULL = "Bearer " + TEST_AZURE_TOKEN
+
 
 class BaseTestCase(unittest.TestCase):
 
@@ -170,7 +178,7 @@ class TestFileOrData(BaseTestCase):
         temp_filename = NON_EXISTING_FILE
         obj = {TEST_FILE_KEY: temp_filename}
         t = FileOrData(obj=obj, file_key_name=TEST_FILE_KEY)
-        self.expect_exception(t.as_file, "does not exists")
+        self.expect_exception(t.as_file, "does not exist")
 
     def test_file_given_data(self):
         obj = {TEST_DATA_KEY: TEST_DATA_BASE64}
@@ -248,6 +256,16 @@ class TestFileOrData(BaseTestCase):
                        data_key_name=TEST_DATA_KEY, base64_file_content=False)
         self.assertEqual(TEST_DATA, self.get_file_content(t.as_file()))
 
+    def test_file_given_no_object(self):
+        t = FileOrData(obj=None, file_key_name=TEST_FILE_KEY,
+                       data_key_name=TEST_DATA_KEY)
+        self.assertEqual(t.as_file(), None)
+
+    def test_file_given_no_object_data(self):
+        t = FileOrData(obj=None, file_key_name=TEST_FILE_KEY,
+                       data_key_name=TEST_DATA_KEY)
+        self.assertEqual(t.as_data(), None)
+
 
 class TestConfigNode(BaseTestCase):
 
@@ -420,6 +438,41 @@ class TestKubeConfigLoader(BaseTestCase):
                     "user": "oidc"
                 }
             },
+            {
+                "name": "azure",
+                "context": {
+                    "cluster": "default",
+                    "user": "azure"
+                }
+            },
+            {
+                "name": "azure_num",
+                "context": {
+                    "cluster": "default",
+                    "user": "azure_num"
+                }
+            },
+            {
+                "name": "azure_str",
+                "context": {
+                    "cluster": "default",
+                    "user": "azure_str"
+                }
+            },
+            {
+                "name": "azure_num_error",
+                "context": {
+                    "cluster": "default",
+                    "user": "azure_str_error"
+                }
+            },
+            {
+                "name": "azure_str_error",
+                "context": {
+                    "cluster": "default",
+                    "user": "azure_str_error"
+                }
+            },
             {
                 "name": "expired_oidc",
                 "context": {
@@ -499,6 +552,34 @@ class TestKubeConfigLoader(BaseTestCase):
                     "user": "exec_cred_user"
                 }
             },
+            {
+                "name": "exec_cred_user_certificate",
+                "context": {
+                    "cluster": "ssl",
+                    "user": "exec_cred_user_certificate"
+                }
+            },
+            {
+                "name": "contexttestcmdpath",
+                "context": {
+                    "cluster": "clustertestcmdpath",
+                    "user": "usertestcmdpath"
+                }
+            },
+            {
+                "name": "contexttestcmdpathempty",
+                "context": {
+                    "cluster": "clustertestcmdpath",
+                    "user": "usertestcmdpathempty"
+                }
+            },
+            {
+                "name": "contexttestcmdpathscope",
+                "context": {
+                    "cluster": "clustertestcmdpath",
+                    "user": "usertestcmdpathscope"
+                }
+            }
         ],
         "clusters": [
             {
@@ -527,15 +608,20 @@ class TestKubeConfigLoader(BaseTestCase):
                     "server": TEST_SSL_HOST,
                     "certificate-authority-data":
                         TEST_CERTIFICATE_AUTH_BASE64,
+                    "insecure-skip-tls-verify": False,
                 }
             },
             {
                 "name": "no_ssl_verification",
                 "cluster": {
                     "server": TEST_SSL_HOST,
-                    "insecure-skip-tls-verify": "true",
+                    "insecure-skip-tls-verify": True,
                 }
             },
+            {
+                "name": "clustertestcmdpath",
+                "cluster": {}
+            }
         ],
         "users": [
             {
@@ -603,6 +689,94 @@ class TestKubeConfigLoader(BaseTestCase):
                     }
                 }
             },
+            {
+                "name": "azure",
+                "user": {
+                    "auth-provider": {
+                        "config": {
+                            "access-token": TEST_AZURE_TOKEN,
+                            "apiserver-id": "00000002-0000-0000-c000-"
+                                            "000000000000",
+                            "environment": "AzurePublicCloud",
+                            "refresh-token": "refreshToken",
+                            "tenant-id": "9d2ac018-e843-4e14-9e2b-4e0ddac75433"
+                        },
+                        "name": "azure"
+                    }
+                }
+            },
+            {
+                "name": "azure_num",
+                "user": {
+                    "auth-provider": {
+                        "config": {
+                            "access-token": TEST_AZURE_TOKEN,
+                            "apiserver-id": "00000002-0000-0000-c000-"
+                                            "000000000000",
+                            "environment": "AzurePublicCloud",
+                            "expires-in": "0",
+                            "expires-on": "156207275",
+                            "refresh-token": "refreshToken",
+                            "tenant-id": "9d2ac018-e843-4e14-9e2b-4e0ddac75433"
+                        },
+                        "name": "azure"
+                    }
+                }
+            },
+            {
+                "name": "azure_str",
+                "user": {
+                    "auth-provider": {
+                        "config": {
+                            "access-token": TEST_AZURE_TOKEN,
+                            "apiserver-id": "00000002-0000-0000-c000-"
+                                            "000000000000",
+                            "environment": "AzurePublicCloud",
+                            "expires-in": "0",
+                            "expires-on": "2018-10-18 00:52:29.044727",
+                            "refresh-token": "refreshToken",
+                            "tenant-id": "9d2ac018-e843-4e14-9e2b-4e0ddac75433"
+                        },
+                        "name": "azure"
+                    }
+                }
+            },
+            {
+                "name": "azure_str_error",
+                "user": {
+                    "auth-provider": {
+                        "config": {
+                            "access-token": TEST_AZURE_TOKEN,
+                            "apiserver-id": "00000002-0000-0000-c000-"
+                                            "000000000000",
+                            "environment": "AzurePublicCloud",
+                            "expires-in": "0",
+                            "expires-on": "2018-10-18 00:52",
+                            "refresh-token": "refreshToken",
+                            "tenant-id": "9d2ac018-e843-4e14-9e2b-4e0ddac75433"
+                        },
+                        "name": "azure"
+                    }
+                }
+            },
+            {
+                "name": "azure_num_error",
+                "user": {
+                    "auth-provider": {
+                        "config": {
+                            "access-token": TEST_AZURE_TOKEN,
+                            "apiserver-id": "00000002-0000-0000-c000-"
+                                            "000000000000",
+                            "environment": "AzurePublicCloud",
+                            "expires-in": "0",
+                            "expires-on": "-1",
+                            "refresh-token": "refreshToken",
+                            "tenant-id": "9d2ac018-e843-4e14-9e2b-4e0ddac75433"
+                        },
+                        "name": "azure"
+                    }
+                }
+            },
             {
                 "name": "expired_oidc",
                 "user": {
@@ -709,6 +883,50 @@ class TestKubeConfigLoader(BaseTestCase):
                     }
                 }
             },
+            {
+                "name": "exec_cred_user_certificate",
+                "user": {
+                    "exec": {
+                        "apiVersion": "client.authentication.k8s.io/v1beta1",
+                        "command": "custom-certificate-authenticator",
+                        "args": []
+                    }
+                }
+            },
+            {
+                "name": "usertestcmdpath",
+                "user": {
+                    "auth-provider": {
+                        "name": "gcp",
+                        "config": {
+                            "cmd-path": "cmdtorun"
+                        }
+                    }
+                }
+            },
+            {
+                "name": "usertestcmdpathempty",
+                "user": {
+                    "auth-provider": {
+                        "name": "gcp",
+                        "config": {
+                            "cmd-path": ""
+                        }
+                    }
+                }
+            },
+            {
+                "name": "usertestcmdpathscope",
+                "user": {
+                    "auth-provider": {
+                        "name": "gcp",
+                        "config": {
+                            "cmd-path": "cmd",
+                            "scopes": "scope"
+                        }
+                    }
+                }
+            }
         ]
     }
 
@@ -886,6 +1104,46 @@ class TestKubeConfigLoader(BaseTestCase):
             None,
         )
 
+    def test_azure_no_refresh(self):
+        loader = KubeConfigLoader(
+            config_dict=self.TEST_KUBE_CONFIG,
+            active_context="azure",
+        )
+        self.assertTrue(loader._load_auth_provider_token())
+        self.assertEqual(TEST_AZURE_TOKEN_FULL, loader.token)
+
+    def test_azure_with_expired_num(self):
+        loader = KubeConfigLoader(
+            config_dict=self.TEST_KUBE_CONFIG,
+            active_context="azure_num",
+        )
+        provider = loader._user['auth-provider']
+        self.assertTrue(loader._azure_is_expired(provider))
+
+    def test_azure_with_expired_str(self):
+        loader = KubeConfigLoader(
+            config_dict=self.TEST_KUBE_CONFIG,
+            active_context="azure_str",
+        )
+        provider = loader._user['auth-provider']
+        self.assertTrue(loader._azure_is_expired(provider))
+
+    def test_azure_with_expired_str_error(self):
+        loader = KubeConfigLoader(
+            config_dict=self.TEST_KUBE_CONFIG,
+            active_context="azure_str_error",
+        )
+        provider = loader._user['auth-provider']
+        self.assertRaises(ValueError, loader._azure_is_expired, provider)
+
+    def test_azure_with_expired_int_error(self):
+        loader = KubeConfigLoader(
+            config_dict=self.TEST_KUBE_CONFIG,
+            active_context="azure_num_error",
+        )
+        provider = loader._user['auth-provider']
+        self.assertRaises(ValueError, loader._azure_is_expired, provider)
+
     def test_user_pass(self):
         expected = FakeConfig(host=TEST_HOST, token=TEST_BASIC_TOKEN)
         actual = FakeConfig()
@@ -907,7 +1165,7 @@ class TestKubeConfigLoader(BaseTestCase):
             active_context="ssl-no_file")
         self.expect_exception(
             loader.load_and_set,
-            "does not exists",
+            "does not exist",
             FakeConfig())
 
     def test_ssl(self):
@@ -916,7 +1174,8 @@ class TestKubeConfigLoader(BaseTestCase):
             token=BEARER_TOKEN_FORMAT % TEST_DATA_BASE64,
             cert_file=self._create_temp_file(TEST_CLIENT_CERT),
             key_file=self._create_temp_file(TEST_CLIENT_KEY),
-            ssl_ca_cert=self._create_temp_file(TEST_CERTIFICATE_AUTH)
+            ssl_ca_cert=self._create_temp_file(TEST_CERTIFICATE_AUTH),
+            verify_ssl=True
         )
         actual = FakeConfig()
         KubeConfigLoader(
@@ -989,7 +1248,7 @@ class TestKubeConfigLoader(BaseTestCase):
         finally:
             shutil.rmtree(temp_dir)
 
-    def test_load_kube_config(self):
+    def test_load_kube_config_from_file_path(self):
         expected = FakeConfig(host=TEST_HOST,
                               token=BEARER_TOKEN_FORMAT % TEST_DATA_BASE64)
         config_file = self._create_temp_file(
@@ -999,6 +1258,76 @@ class TestKubeConfigLoader(BaseTestCase):
                          client_configuration=actual)
         self.assertEqual(expected, actual)
 
+    def test_load_kube_config_from_file_like_object(self):
+        expected = FakeConfig(host=TEST_HOST,
+                              token=BEARER_TOKEN_FORMAT % TEST_DATA_BASE64)
+        config_file_like_object = io.StringIO()
+        # py3 (won't have unicode) vs py2 (requires it)
+        try:
+            unicode('')
+            config_file_like_object.write(
+                unicode(
+                    yaml.safe_dump(
+                        self.TEST_KUBE_CONFIG),
+                    errors='replace'))
+        except NameError:
+            config_file_like_object.write(
+                yaml.safe_dump(
+                    self.TEST_KUBE_CONFIG))
+        actual = FakeConfig()
+        load_kube_config(
+            config_file=config_file_like_object,
+            context="simple_token",
+            client_configuration=actual)
+        self.assertEqual(expected, actual)
+
+    def test_load_kube_config_from_dict(self):
+        expected = FakeConfig(host=TEST_HOST,
+                              token=BEARER_TOKEN_FORMAT % TEST_DATA_BASE64)
+        actual = FakeConfig()
+        load_kube_config_from_dict(config_dict=self.TEST_KUBE_CONFIG,
+                                   context="simple_token",
+                                   client_configuration=actual)
+        self.assertEqual(expected, actual)
+
+    def test_load_kube_config_from_dict_with_temp_file_path(self):
+        expected = FakeConfig(
+            host=TEST_SSL_HOST,
+            token=BEARER_TOKEN_FORMAT % TEST_DATA_BASE64,
+            cert_file=self._create_temp_file(TEST_CLIENT_CERT),
+            key_file=self._create_temp_file(TEST_CLIENT_KEY),
+            ssl_ca_cert=self._create_temp_file(TEST_CERTIFICATE_AUTH),
+            verify_ssl=True
+        )
+        actual = FakeConfig()
+        tmp_path = os.path.join(
+            os.path.dirname(
+                os.path.dirname(
+                    os.path.abspath(__file__))),
+            'tmp_file_path_test')
+        load_kube_config_from_dict(config_dict=self.TEST_KUBE_CONFIG,
+                                   context="ssl",
+                                   client_configuration=actual,
+                                   temp_file_path=tmp_path)
+        self.assertFalse(True if not os.listdir(tmp_path) else False)
+        self.assertEqual(expected, actual)
+        _cleanup_temp_files
+
+    def test_load_kube_config_from_empty_file_like_object(self):
+        config_file_like_object = io.StringIO()
+        self.assertRaises(
+            ConfigException,
+            load_kube_config,
+            config_file_like_object)
+
+    def test_load_kube_config_from_empty_file(self):
+        config_file = self._create_temp_file(
+            yaml.safe_dump(None))
+        self.assertRaises(
+            ConfigException,
+            load_kube_config,
+            config_file)
+
     def test_list_kube_config_contexts(self):
         config_file = self._create_temp_file(
             yaml.safe_dump(self.TEST_KUBE_CONFIG))
@@ -1054,6 +1383,109 @@ class TestKubeConfigLoader(BaseTestCase):
             active_context="exec_cred_user").load_and_set(actual)
         self.assertEqual(expected, actual)
 
+    @mock.patch('kubernetes.config.kube_config.ExecProvider.run')
+    def test_user_exec_auth_certificates(self, mock):
+        mock.return_value = {
+            "clientCertificateData": TEST_CLIENT_CERT,
+            "clientKeyData": TEST_CLIENT_KEY,
+        }
+        expected = FakeConfig(
+            host=TEST_SSL_HOST,
+            cert_file=self._create_temp_file(TEST_CLIENT_CERT),
+            key_file=self._create_temp_file(TEST_CLIENT_KEY),
+            ssl_ca_cert=self._create_temp_file(TEST_CERTIFICATE_AUTH),
+            verify_ssl=True)
+        actual = FakeConfig()
+        KubeConfigLoader(
+            config_dict=self.TEST_KUBE_CONFIG,
+            active_context="exec_cred_user_certificate").load_and_set(actual)
+        self.assertEqual(expected, actual)
+
+    def test_user_cmd_path(self):
+        A = namedtuple('A', ['token', 'expiry'])
+        token = "dummy"
+        return_value = A(token, parse_rfc3339(datetime.datetime.now()))
+        CommandTokenSource.token = mock.Mock(return_value=return_value)
+        expected = FakeConfig(api_key={
+                              "authorization": BEARER_TOKEN_FORMAT % token})
+        actual = FakeConfig()
+        KubeConfigLoader(
+            config_dict=self.TEST_KUBE_CONFIG,
+            active_context="contexttestcmdpath").load_and_set(actual)
+        del actual.get_api_key_with_prefix
+        self.assertEqual(expected, actual)
+
+    def test_user_cmd_path_empty(self):
+        A = namedtuple('A', ['token', 'expiry'])
+        token = "dummy"
+        return_value = A(token, parse_rfc3339(datetime.datetime.now()))
+        CommandTokenSource.token = mock.Mock(return_value=return_value)
+        expected = FakeConfig(api_key={
+                              "authorization": BEARER_TOKEN_FORMAT % token})
+        actual = FakeConfig()
+        self.expect_exception(lambda: KubeConfigLoader(
+            config_dict=self.TEST_KUBE_CONFIG,
+            active_context="contexttestcmdpathempty").load_and_set(actual),
+            "missing access token cmd "
+            "(cmd-path is an empty string in your kubeconfig file)")
+
+    def test_user_cmd_path_with_scope(self):
+        A = namedtuple('A', ['token', 'expiry'])
+        token = "dummy"
+        return_value = A(token, parse_rfc3339(datetime.datetime.now()))
+        CommandTokenSource.token = mock.Mock(return_value=return_value)
+        expected = FakeConfig(api_key={
+                              "authorization": BEARER_TOKEN_FORMAT % token})
+        actual = FakeConfig()
+        self.expect_exception(lambda: KubeConfigLoader(
+            config_dict=self.TEST_KUBE_CONFIG,
+            active_context="contexttestcmdpathscope").load_and_set(actual),
+            "scopes can only be used when kubectl is using "
+            "a gcp service account key")
+
+    def test__get_kube_config_loader_for_yaml_file_no_persist(self):
+        expected = FakeConfig(host=TEST_HOST,
+                              token=BEARER_TOKEN_FORMAT % TEST_DATA_BASE64)
+        config_file = self._create_temp_file(
+            yaml.safe_dump(self.TEST_KUBE_CONFIG))
+        actual = _get_kube_config_loader_for_yaml_file(config_file)
+        self.assertIsNone(actual._config_persister)
+
+    def test__get_kube_config_loader_for_yaml_file_persist(self):
+        expected = FakeConfig(host=TEST_HOST,
+                              token=BEARER_TOKEN_FORMAT % TEST_DATA_BASE64)
+        config_file = self._create_temp_file(
+            yaml.safe_dump(self.TEST_KUBE_CONFIG))
+        actual = _get_kube_config_loader_for_yaml_file(config_file,
+                                                       persist_config=True)
+        self.assertTrue(callable(actual._config_persister))
+        self.assertEqual(actual._config_persister.__name__, "save_changes")
+
+    def test__get_kube_config_loader_file_no_persist(self):
+        expected = FakeConfig(host=TEST_HOST,
+                              token=BEARER_TOKEN_FORMAT % TEST_DATA_BASE64)
+        config_file = self._create_temp_file(
+            yaml.safe_dump(self.TEST_KUBE_CONFIG))
+        actual = _get_kube_config_loader(filename=config_file)
+        self.assertIsNone(actual._config_persister)
+
+    def test__get_kube_config_loader_file_persist(self):
+        expected = FakeConfig(host=TEST_HOST,
+                              token=BEARER_TOKEN_FORMAT % TEST_DATA_BASE64)
+        config_file = self._create_temp_file(
+            yaml.safe_dump(self.TEST_KUBE_CONFIG))
+        actual = _get_kube_config_loader(filename=config_file,
+                                         persist_config=True)
+        self.assertTrue(callable(actual._config_persister))
+        self.assertEquals(actual._config_persister.__name__, "save_changes")
+
+    def test__get_kube_config_loader_dict_no_persist(self):
+        expected = FakeConfig(host=TEST_HOST,
+                              token=BEARER_TOKEN_FORMAT % TEST_DATA_BASE64)
+        actual = _get_kube_config_loader(
+            config_dict=self.TEST_KUBE_CONFIG)
+        self.assertIsNone(actual._config_persister)
+
 
 class TestKubernetesClientConfiguration(BaseTestCase):
     # Verifies properties of kubernetes.client.Configuration.
@@ -1073,11 +1505,13 @@ class TestKubernetesClientConfiguration(BaseTestCase):
 
     def test_auth_settings_calls_get_api_key_with_prefix(self):
         expected_token = 'expected_token'
+        old_token = 'old_token'
 
         def fake_get_api_key_with_prefix(identifier):
             self.assertEqual('authorization', identifier)
             return expected_token
         config = Configuration()
+        config.api_key['authorization'] = old_token
         config.get_api_key_with_prefix = fake_get_api_key_with_prefix
         self.assertEqual(expected_token,
                          config.auth_settings()['BearerToken']['value'])
@@ -1193,13 +1627,70 @@ class TestKubeConfigMerger(BaseTestCase):
             },
         ]
     }
+    TEST_KUBE_CONFIG_PART4 = {
+        "current-context": "no_user",
+    }
+    # Config with user having cmd-path
+    TEST_KUBE_CONFIG_PART5 = {
+        "contexts": [
+            {
+                "name": "contexttestcmdpath",
+                "context": {
+                    "cluster": "clustertestcmdpath",
+                    "user": "usertestcmdpath"
+                }
+            }
+        ],
+        "clusters": [
+            {
+                "name": "clustertestcmdpath",
+                "cluster": {}
+            }
+        ],
+        "users": [
+            {
+                "name": "usertestcmdpath",
+                "user": {
+                    "auth-provider": {
+                        "name": "gcp",
+                        "config": {
+                            "cmd-path": "cmdtorun"
+                        }
+                    }
+                }
+            }
+        ]
+    }
+    TEST_KUBE_CONFIG_PART6 = {
+        "current-context": "no_user",
+        "contexts": [
+            {
+                "name": "no_user",
+                "context": {
+                    "cluster": "default"
+                }
+            },
+        ],
+        "clusters": [
+            {
+                "name": "default",
+                "cluster": {
+                    "server": TEST_HOST
+                }
+            },
+        ],
+        "users": None
+    }
 
     def _create_multi_config(self):
         files = []
         for part in (
                 self.TEST_KUBE_CONFIG_PART1,
                 self.TEST_KUBE_CONFIG_PART2,
-                self.TEST_KUBE_CONFIG_PART3):
+                self.TEST_KUBE_CONFIG_PART3,
+                self.TEST_KUBE_CONFIG_PART4,
+                self.TEST_KUBE_CONFIG_PART5,
+                self.TEST_KUBE_CONFIG_PART6):
             files.append(self._create_temp_file(yaml.safe_dump(part)))
         return ENV_KUBECONFIG_PATH_SEPARATOR.join(files)
 
@@ -1210,7 +1701,11 @@ class TestKubeConfigMerger(BaseTestCase):
             {'context': {'cluster': 'ssl', 'user': 'ssl'}, 'name': 'ssl'},
             {'context': {'cluster': 'default', 'user': 'simple_token'},
              'name': 'simple_token'},
-            {'context': {'cluster': 'default', 'user': 'expired_oidc'}, 'name': 'expired_oidc'}]
+            {'context': {'cluster': 'default', 'user': 'expired_oidc'},
+             'name': 'expired_oidc'},
+            {'context': {'cluster': 'clustertestcmdpath',
+                         'user': 'usertestcmdpath'},
+             'name': 'contexttestcmdpath'}]
 
         contexts, active_context = list_kube_config_contexts(
             config_file=kubeconfigs)
